
#include "base62.h"

typedef unsigned long long uint64_t_;

static inline int b62enc(int c)
{
	return "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"[c];
}

static inline int b62dec(int c)
{
	static unsigned char table[] = {
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,0xff,
		0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0xff,0xff,0xff,0xff,0xff,0xff,
		0xff,0x0a,0x0b,0x0c,0x0d,0x0e,0x0f,0x10,0x11,0x12,0x13,0x14,0x15,0x16,0x17,0x18,
		0x19,0x1a,0x1b,0x1c,0x1d,0x1e,0x1f,0x20,0x21,0x22,0x23,0xff,0xff,0xff,0xff,0xff,
		0xff,0x24,0x25,0x26,0x27,0x28,0x29,0x2a,0x2b,0x2c,0x2d,0x2e,0x2f,0x30,0x31,0x32,
		0x33,0x34,0x35,0x36,0x37,0x38,0x39,0x3a,0x3b,0x3c,0x3d,0xff,0xff,0xff,0xff,0xff,
	};
	return (c >= 0 && c < 0x80) ? table[c] : 0xff;
}

std::string base62_encode(unsigned char const *ptr, size_t len)
{
	std::vector<char> out;
	{
		int l = ((len + 55) / 56) * 78;
		out.reserve(l);
	}
	int pos = 0;
	while (pos < (int)len) {
		int l = len;
		if (l > 56) {
			l = 56;
		}
		out.push_back(b62enc(l));
		for (int i = 0; i < l; i += 8) {
			uint64_t_ data = 0;
			for (int j = 0; j < 8; j++) {
				int c = (i + j) < l ? ptr[pos + i + j] : 0;
				data |= (uint64_t_)c << ((j & 7) * 8);
			}
			for (int j = 0; j < 11; j++) {
				int d = data % 62;
				data /= 62;
				out.push_back(b62enc(d));
			}
		}
		pos += l;
	}
	if (out.empty()) return std::string();
	return std::string(&out[0], out.size());
}

bool base62_decode(char const *ptr, size_t len, std::vector<unsigned char> *out)
{
	out->clear();
	{
		int l = 0;
		int i = 0;
		while (i < (int)len) {
			int n = b62dec(ptr[i]);
			if (n == 0xff) return false;
			if (n == 0) break;
			l += n;
			n = ((n + 7) / 8) * 11 + 1;
			i += n;
		}
		out->reserve(l);
	}
	int pos = 0;
	while (pos < (int)len) {
		int l = b62dec(ptr[pos]);
		if (l == 0xff) return false;
		if (l == 0) break;
		pos++;
		for (int i = 0; i < l; i += 8) {
			if (len < 11) return false;
			uint64_t_ data = 0;
			for (int j = 0; j < 11; j++) {
				int n = b62dec(ptr[pos + 10 - j]);
				if (n == 0xff) return false;
				data = data * 62 + n;
			}
			int m = l - i;
			if (m > 8) {
				m = 8;
			}
			for (int j = 0; j < m; j++) {
				out->push_back((unsigned char)(data >> ((j & 7) * 8)));
			}
			pos += 11;
		}
	}
	return true;
}

